#include <cmath>
#include<gsl/gsl_sf_erf.h>
//! Call premium
double Call(double S, double K, double r, double Vol, double theta)
  //!S : Stock price --- K : Strike --- r: riskless interest rate 
  //!Vol: volatility --- theta: time to maturity
{
  if(S>0)
    {
      double standard_deviation= Vol*sqrt(theta);
      double d1 = (log(S/K)+r*theta)/standard_deviation 
	           + 0.5*standard_deviation;
      double d2 = d1 -  standard_deviation;
      return S*gsl_sf_erf_Q(-d1) - K*exp(-r*theta)* gsl_sf_erf_Q(-d2);
    }
  else
    return 0;
}

//! Put premium
double Put(double S, double K, double r, double Vol, double theta)
  //!S : Stock price --- K : Strike --- r: riskless interest rate 
  //!Vol: volatility --- theta: time to maturity
{
  if (S>0)
    {
      double standard_deviation= Vol*sqrt(theta);
      double d1 = (log(S/K)+r*theta)/standard_deviation + 0.5*standard_deviation;
      double d2 = d1 -  standard_deviation;
      return -S*gsl_sf_erf_Q(d1)+K*exp(-r*theta)* gsl_sf_erf_Q(d2);
    }
  else
    return K*exp(-r*theta);
}

