#define CHECK_KN
#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <sstream>
#include <string>

#include "RefCounter.hpp"
#include "RNM.hpp"
#include "euler_scheme_RNM.hpp"

#include "adaption.hpp"
#include "BS_formula.hpp"

#define CHGT_S2Y(s)   ((s)/(1.+(s)))
#define CHGT_Y2S(y)   ((y)/(1.-(y)))


using namespace std;

#if 0
#define SQUARE(x) pow((x),2)
#else
#define SQUARE(x) ((x)*(x))
#endif

double vol(double t,double S)
{
#if 0
  // return 0.2;
  // if (max(abs(t-0.5)/0.1, abs(S-100)/20.0) <=1)
  if (SQUARE(t-0.5)/0.01+SQUARE(S-100)/400<=1)
    return 0.3;
  else
    if (SQUARE(t-0.9)/0.01+SQUARE(S-100)/400<=1)
      return 0.3;
    else
      return 0.05;
#else
      return 0.2;
#endif
}


void InternalError(const char * str) {cerr << str; exit(1);}

int main(void)
{
  int Nt=21;
  int NS=81;
  vector< vector<double> >grid_t(60);
  vector<vector<vector<double> > >grid_S(60); // grid_S[a][b][c] -> a = n. de grille, b=temps, c=S
  vector< vector<vector<double> > >S_steps(60);
  vector < vector<int> >change_grid(60);
  vector <double> en_1;
  vector <double> en_2;

  double S_min = 10., S_max= 200., T=1.;
  double K=100;

  grid_S[0].resize(Nt);
  S_steps[0].resize(Nt);
  change_grid[0].resize(Nt,0);
  
  for(int i=0;i<Nt;i++)
    {
      grid_t[0].push_back((T*i)/(Nt-1));
      for(int j=0;j<NS;j++)
        grid_S[ 0 ][ i ].push_back(CHGT_S2Y(S_min+(S_max-S_min)*j/(NS-1)));
      for(int j=0;j<NS-1;j++)
	S_steps[0][i].push_back(grid_S[0][i][j+1]-grid_S[0][i][j]);
    }
  int ref_it=0;
  double  estim_norm_error=10000;
  while(ref_it<20 &&  estim_norm_error>1)
    {
    
      Euler_Scheme scheme(grid_S[ref_it], S_steps[ref_it],grid_t[ref_it],change_grid[ref_it]);
      vector<KN<double> >  P(Nt);
      vector<KN<double> >  error(Nt);
      vector<KN<double> >  estim_S(Nt);
      for (int i=0;i<P.size();i++)
	{
	  P[i].init(S_steps[ref_it][i].size());
	  error[i].init(S_steps[ref_it][i].size());
	  estim_S[i].init(S_steps[ref_it][i].size());
	  for(int j=0; j<S_steps[ref_it][i].size();j++)
	  {
	    P[i](j)=0;
	    estim_S[i](j)=0;
	    error[i](j)=0;
	  }
	}      
      int  iS=0;    
      while (CHGT_Y2S(grid_S[ref_it][0][iS])<K)
	{
	  P[0](iS)=K-CHGT_Y2S(grid_S[ref_it][0][iS]);
	  iS++;
	}
      vector<double> estim;
      vector<double> estim_S_global;
      vector<double> estim_global;
      vector<double> fb;
      int free_bdry=iS;
      fb.push_back(K);
      double norm_estim=0, norm_estim_S_global=0;
      for(int i=1;i<Nt;i++)
	{
	  scheme.Time_Step(i,P);
	  double aux1=sqrt(scheme.build_time_error_estimator(i,P))* vol(0.5,100);
	  estim.push_back(aux1);   //indicateur d erreur en temps
	  norm_estim+=aux1*aux1;
	  //	  cout<< "i="<<i<< "indic="<< estim[i-1] <<endl;
	  scheme.build_S_estimator_1(i,P,estim_S);          //indicateur d erreur en S 
	  double aux=scheme.build_global_S_estimator(i,estim_S[i],grid_S[ref_it][i]);
	  estim_S_global.push_back(aux); 
	  norm_estim_S_global+=aux*aux;
	  estim_global.push_back(aux+aux1); 
	}
      
      

      for(int i=1;i<grid_t[ref_it].size();i++)
	for(int j=0;j<grid_S[ref_it][i].size()-1;j++)
	  error[i](j)=P[i](j)-Put( CHGT_Y2S(grid_S[ref_it][i][j]),K,0.04,0.2, grid_t[ref_it][i]);
      double energy=0;
      scheme.energy(error,energy);
      energy=sqrt(energy);

      cout << "refiter=" <<ref_it<< " energy_error="<<energy<<" norm estim ="<< sqrt( norm_estim)<< "  norm_estim_S_global="<<  sqrt( norm_estim_S_global)<<endl;
      en_1.push_back(energy);

      estim_norm_error=sqrt( norm_estim)+ sqrt( norm_estim_S_global);
      en_2.push_back( estim_norm_error);

      std::ostringstream sols;
      sols<<"solution"<< ref_it;
      ofstream outf(sols.str().c_str());       
      for(int i=1;i<grid_t[ref_it].size()-1;i++)
	{
	  int k=0;
	  for(int j=1;j<grid_S[ref_it][i].size()-2;j++)
	    {
	      double Sl=CHGT_Y2S(grid_S[ref_it][i][j]);
	      outf<<  Sl<<"\t"<<grid_t[ref_it][i]<<"\t"<</*fabs(*/P[i](j)/* - Put(Sl,K,0.04,0.2, grid_t[ref_it][i]))*/<<endl;
	      double Sr=CHGT_Y2S(grid_S[ref_it][i][j+1]);
	      outf<< Sr<<"\t"<<grid_t[ref_it][i]<<"\t"<< /*fabs(*/P[i](j+1)/* - Put(Sr,K,0.04,0.2, grid_t[ref_it][i]))*/<<endl;
	      if (change_grid[ref_it][i]==0)	 
		{
		  outf<< Sr<<"\t"<<grid_t[ref_it][i-1]<<"\t"<</*fabs(*/P[i-1](j+1)/* - Put(Sr,K,0.04,0.2, grid_t[ref_it][i-1]))*/<<endl;
		  outf<< Sl<<"\t"<<grid_t[ref_it][i-1]<<"\t"<</*fabs(*/P[i-1](j)/* - Put(Sl,K,0.04,0.2, grid_t[ref_it][i-1]))*/<<endl;
		  outf<< Sl<<"\t"<<grid_t[ref_it][i]<<"\t"<< /*fabs(*/P[i](j)/* - Put(Sl,K,0.04,0.2, grid_t[ref_it][i]))*/<<endl;
		  outf<<endl;
		}
	      else
		{   
		  while (CHGT_Y2S( grid_S[ref_it][i-1][k])<Sl)
		    k++;
		  int kl=k-1;
		  double Tmpk = CHGT_Y2S( grid_S[ ref_it ][ i-1 ][ k ] );
		  double Tmpkl = CHGT_Y2S( grid_S[ ref_it ][ i-1 ][ kl ] );
		  double xl=(Sl-Tmpkl)/(Tmpk-Tmpkl);
		  while (Tmpk<Sr)
		    k++;
		  int kr=k-1;
		  double Tmpkr = CHGT_Y2S( grid_S[ref_it][i-1][kr] );
		  double xr=(Sr-Tmpkr)/(Tmpk-Tmpkr);
		  outf<< Sr<<"\t"<<grid_t[ref_it][i-1]<<"\t"<</*fabs(*/P[i-1](kr)*(1-xr)+ P[i-1](kr+1)*xr/* - Put(Sr,K,0.04,0.2, grid_t[ref_it][i-1]))*/<<endl;
		  outf<< Sl<<"\t"<<grid_t[ref_it][i-1]<<"\t"<</*fabs(*/P[i-1](kl)*(1-xl)+ P[i-1](kl+1)*xl/* - Put(Sl,K,0.04,0.2, grid_t[ref_it][i-1]))*/<<endl;
		  outf<< Sl<<"\t"<<grid_t[ref_it][i]<<"\t"<</*fabs(*/P[i](j)/* - Put(Sl,K,0.04,0.2, grid_t[ref_it][i]))*/<<endl;
		  outf<<endl;
		}
	    }
	  outf<<endl;
	} 

      std::ostringstream indics;
      indics<<"indic"<< ref_it;
      ofstream oute(indics.str().c_str());       
      for(int i=1;i<Nt;i++)
	oute<<grid_t[ref_it][i]<<"\t"<< estim[i-1]<<endl;
      std::ostringstream indicglobs;
      indicglobs<<"indicsglob"<< ref_it;
      ofstream outeglob(indicglobs.str().c_str());          
      for(int i=1;i<Nt;i++)
	outeglob<<grid_t[ref_it][i]<<"\t"<< estim_S_global[i-1]<<endl;
  
      std::ostringstream indic_Ss;
      indic_Ss<<"indic_S"<< ref_it;
      ofstream outes(indic_Ss.str().c_str());      

      for(int i=1;i<Nt;i++)
	{
	  for(int j=0;j<grid_S[ref_it][i].size()-2;j++)
	    outes<< CHGT_Y2S( grid_S[ref_it][i][j] )<<"\t"<<grid_t[ref_it][i]<<"\t"<<estim_S[i](j)<<endl;
	  outes<<endl;
	}
      int max_aspect_ratio_t=2;
      int max_aspect_ratio=2;
      double reduction_factor=8.;

      int which_refine=ref_it%2;
      if ( norm_estim/norm_estim_S_global>2|| norm_estim/norm_estim_S_global<0.5)
	which_refine=norm_estim<norm_estim_S_global;
      if (which_refine==0)
	cout << "raffinement a "<< ref_it <<" en t"<<endl;
      else
	cout << "raffinement a "<< ref_it <<" en S"<<endl;
      adaption_tS(grid_t[ref_it],estim,grid_t[ref_it+1],1,max_aspect_ratio_t ,grid_S[ref_it], change_grid[ref_it], S_steps[ref_it],
		  estim_S, grid_S[ref_it+1],S_steps[ref_it+1], change_grid[ref_it+1], reduction_factor, max_aspect_ratio,  estim_S_global,
		  	which_refine);
      //ref_it%2);
      Nt=grid_t[ref_it+1].size();
      ref_it++;   




      std::ostringstream indicglob;
      indicglob<<"indicglob"<< ref_it;
      ofstream outg(indicglob.str().c_str());       
      for(int i=1;i<Nt;i++)
	outg<<grid_t[ref_it][i]<<"\t"<< estim_global[i-1]<<endl;
    }

  ofstream outcomp("error_1");       
  for(int i=0;i<en_1.size();i++)
    outcomp<< i<<"\t"<<en_1[i]<<endl;
  ofstream outcomp2("error_2");       
  for(int i=0;i<en_2.size();i++)
    outcomp2<< i<<"\t"<<en_2[i]<<endl;
  
}
